from typing import Any, AsyncGenerator, List, Optional

from promptl_ai import Message

from latitude_sdk.client import Client, ClientEvent, RequestHandler, ToolResultsRequestBody
from latitude_sdk.sdk.errors import ApiError, ApiErrorCodes
from latitude_sdk.sdk.types import (
    ChainEvents,
    FinishedResult,
    OnToolCall,
    OnToolCallDetails,
    ProviderEvents,
    ProviderEventToolCalled,
    StreamCallbacks,
    StreamEvents,
    ToolResult,
    _LatitudeEvent,
)


class Streamer:
    _client: Client

    def __init__(self, client: Client):
        self._client = client

    async def handle(
        self,
        stream: AsyncGenerator[ClientEvent, Any],
        on_event: Optional[StreamCallbacks.OnEvent],
        tools: Optional[dict[str, OnToolCall]],
    ) -> FinishedResult:
        uuid = None
        conversation: List[Message] = []
        response = None

        async for stream_event in stream:
            event = None
            tool_call = None

            if stream_event.event == str(StreamEvents.Latitude):
                event = _LatitudeEvent.validate_json(stream_event.data)
                conversation = event.messages
                uuid = event.uuid

                if event.type == ChainEvents.ProviderCompleted:
                    response = event.response

                elif event.type == ChainEvents.ChainError:
                    raise ApiError(
                        status=400,
                        code=ApiErrorCodes.AIRunError,
                        message=event.error.message,
                        response=stream_event.data,
                    )

            elif stream_event.event == str(StreamEvents.Provider):
                event = stream_event.json()
                event["event"] = StreamEvents.Provider

                if event.get("type") == str(ProviderEvents.ToolCalled):
                    tool_call = ProviderEventToolCalled.model_validate_json(stream_event.data)

            else:
                raise ApiError(
                    status=500,
                    code=ApiErrorCodes.InternalServerError,
                    message=f"Unknown stream event: {stream_event.event}",
                    response=stream_event.data,
                )

            if on_event:
                await on_event(event)

            if tool_call:
                await self._handle_tool_call(tool_call, tools)

        if not uuid or not response:
            raise ApiError(
                status=500,
                code=ApiErrorCodes.InternalServerError,
                message="Stream ended without a chain-complete event. Missing uuid or response.",
                response="Stream ended without a chain-complete event. Missing uuid or response.",
            )

        return FinishedResult(uuid=uuid, conversation=conversation, response=response)

    @staticmethod
    async def _wrap_tool_handler(
        handler: OnToolCall, arguments: dict[str, Any], details: OnToolCallDetails
    ) -> ToolResult:
        tool_result: dict[str, Any] = {"id": details.id, "name": details.name}

        try:
            result = await handler(arguments, details)

            return ToolResult(**tool_result, result=result, is_error=False)
        except Exception as exception:
            return ToolResult(**tool_result, result=str(exception), is_error=True)

    async def _handle_tool_call(
        self, tool_call: ProviderEventToolCalled, tools: Optional[dict[str, OnToolCall]]
    ) -> None:
        # NOTE: Do not handle tool calls if user specified no tools
        if not tools:
            return

        tool_handler = tools.get(tool_call.name)
        # NOTE: If handler not found, do not handle tool call because it could be a built-in tool
        if not tool_handler:
            return

        tool_result = await self._wrap_tool_handler(
            tool_handler,
            tool_call.arguments,
            OnToolCallDetails(
                id=tool_call.id,
                name=tool_call.name,
                arguments=tool_call.arguments,
            ),
        )

        async with self._client.request(
            handler=RequestHandler.ToolResults,
            body=ToolResultsRequestBody(
                tool_call_id=tool_call.id,
                result=tool_result.result,
                is_error=tool_result.is_error,
            ),
        ):
            return



import asyncio
import json
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Optional

import httpx
import httpx_sse

from latitude_sdk.client.payloads import ErrorResponse, RequestBody, RequestHandler, RequestParams
from latitude_sdk.client.router import Router, RouterOptions
from latitude_sdk.sdk.errors import ApiError, ApiErrorCodes, ApiErrorDbRef
from latitude_sdk.sdk.types import LogSources
from latitude_sdk.util import Model
from latitude_sdk.version import version

RETRIABLE_STATUSES = [408, 429, 500, 502, 503, 504]

ClientEvent = httpx_sse.ServerSentEvent


class ClientResponse(httpx.Response):
    async def sse(self: httpx.Response) -> AsyncGenerator[ClientEvent, Any]:
        source = httpx_sse.EventSource(self)

        async for event in source.aiter_sse():
            yield event


httpx.Response.sse = ClientResponse.sse  # pyright: ignore [reportAttributeAccessIssue]


class ClientOptions(Model):
    api_key: str
    retries: int
    delay: float
    timeout: Optional[float]
    source: LogSources
    router: RouterOptions


class Client:
    options: ClientOptions
    router: Router

    def __init__(self, options: ClientOptions):
        self.options = options
        self.router = Router(options.router)

    @asynccontextmanager
    async def request(
        self,
        handler: RequestHandler,
        params: Optional[RequestParams] = None,
        body: Optional[RequestBody] = None,
        stream: Optional[bool] = None,
    ) -> AsyncGenerator[ClientResponse, Any]:
        client = httpx.AsyncClient(
            headers={
                "Authorization": f"Bearer {self.options.api_key}",
                "X-Latitude-SDK-Version": version.semver,
                "Content-Type": "application/json",
                "Accept": "text/event-stream" if stream else "application/json",
            },
            timeout=self.options.timeout,
            follow_redirects=False,
            max_redirects=0,
        )
        response = None
        attempt = 1

        try:
            method, url = self.router.resolve(handler, params)
            content = None
            if body:
                content = json.dumps(
                    {
                        **json.loads(body.model_dump_json()),
                        "__internal": {"source": self.options.source},
                    }
                )

            while attempt <= self.options.retries:
                try:
                    request = client.build_request(method=method, url=url, content=content)
                    response = await client.send(request=request, stream=stream or False)
                    response.raise_for_status()

                    yield response  # pyright: ignore [reportReturnType]
                    break

                except Exception as exception:
                    if isinstance(exception, ApiError):
                        raise exception

                    if attempt >= self.options.retries:
                        raise await self._exception(exception, response) from exception

                    if response and response.status_code in RETRIABLE_STATUSES:
                        await asyncio.sleep(self.options.delay * (2 ** (attempt - 1)))
                    else:
                        raise await self._exception(exception, response) from exception

                finally:
                    if response:
                        await response.aclose()

                    attempt += 1

        except Exception as exception:
            if isinstance(exception, ApiError):
                raise exception

            raise await self._exception(exception, response) from exception

        finally:
            await client.aclose()

    async def _exception(self, exception: Exception, response: Optional[httpx.Response] = None) -> ApiError:
        if not response or not response.is_error:
            return ApiError(
                status=500,
                code=ApiErrorCodes.InternalServerError,
                message=str(exception),
                response=str(exception),
            )

        try:
            if not response.is_stream_consumed and not response.is_closed:
                await response.aread()

            error = ErrorResponse.model_validate_json(response.content)

            return ApiError(
                status=response.status_code,
                code=error.code,
                message=error.message,
                response=response.text,
                db_ref=ApiErrorDbRef(**dict(error.db_ref)) if error.db_ref else None,
            )

        except httpx.ResponseNotRead:
            return ApiError(
                status=500,
                code=ApiErrorCodes.InternalServerError,
                message=str(exception),
                response=str(exception),
            )

        except Exception:
            return ApiError(
                status=response.status_code,
                code=ApiErrorCodes.InternalServerError,
                message=str(exception),
                response=response.text,
            )



from typing import Optional

from latitude_sdk.client import (
    AttachRunRequestBody,
    AttachRunRequestParams,
    Client,
    RequestHandler,
    StopRunRequestParams,
)
from latitude_sdk.sdk.errors import ApiError, ApiErrorCodes
from latitude_sdk.sdk.stream import Streamer
from latitude_sdk.sdk.types import (
    FinishedResult,
    OnToolCall,
    SdkOptions,
    StreamCallbacks,
)
from latitude_sdk.util import Model


class AttachRunOptions(StreamCallbacks, Model):
    tools: Optional[dict[str, OnToolCall]] = None
    stream: Optional[bool] = True  # Note: making stream the default


class AttachRunResult(FinishedResult, Model):
    pass


class Runs:
    _options: SdkOptions
    _client: Client
    _streamer: Streamer

    def __init__(self, client: Client, streamer: Streamer, options: SdkOptions):
        self._options = options
        self._client = client
        self._streamer = streamer

    async def attach(self, uuid: str, options: Optional[AttachRunOptions] = None) -> Optional[AttachRunResult]:
        options = AttachRunOptions(**{**dict(self._options), **dict(options or {})})

        try:
            async with self._client.request(
                handler=RequestHandler.AttachRun,
                params=AttachRunRequestParams(
                    conversation_uuid=uuid,
                ),
                body=AttachRunRequestBody(
                    stream=options.stream,
                ),
                stream=options.stream,
            ) as response:
                if options.stream:
                    result = await self._streamer.handle(response.sse(), options.on_event, options.tools)
                else:
                    result = FinishedResult.model_validate_json(response.content)

            if options.on_finished:
                await options.on_finished(result)

            return AttachRunResult(**dict(result))

        except Exception as exception:
            if not isinstance(exception, ApiError):
                exception = ApiError(
                    status=500,
                    code=ApiErrorCodes.InternalServerError,
                    message=str(exception),
                    response=str(exception),
                )

            if not options.on_error:
                raise exception

            await options.on_error(exception)

            return None

    async def stop(self, uuid: str) -> None:
        async with self._client.request(
            handler=RequestHandler.StopRun,
            params=StopRunRequestParams(
                conversation_uuid=uuid,
            ),
        ):
            return None



from typing import List

from latitude_sdk.client import Client, CreateProjectRequestBody, GetAllVersionsRequestParams, RequestHandler
from latitude_sdk.sdk.types import Project, SdkOptions, Version
from latitude_sdk.util import Adapter as AdapterUtil
from latitude_sdk.util import Model

_GetAllProjectsResult = AdapterUtil[List[Project]](List[Project])
_GetAllVersionsResult = AdapterUtil[List[Version]](List[Version])


class CreateProjectResult(Model):
    project: Project
    version: Version


class Projects:
    _options: SdkOptions
    _client: Client

    def __init__(self, client: Client, options: SdkOptions):
        self._options = options
        self._client = client

    async def get_all(self) -> List[Project]:
        async with self._client.request(
            handler=RequestHandler.GetAllProjects,
        ) as response:
            return _GetAllProjectsResult.validate_json(response.content)

    async def create(self, name: str) -> CreateProjectResult:
        async with self._client.request(
            handler=RequestHandler.CreateProject,
            body=CreateProjectRequestBody(name=name),
        ) as response:
            return CreateProjectResult.model_validate_json(response.content)

    async def get_all_versions(self, project_id: int) -> List[Version]:
        async with self._client.request(
            handler=RequestHandler.GetAllVersions,
            params=GetAllVersionsRequestParams(project_id=project_id),
        ) as response:
            return _GetAllVersionsResult.validate_json(response.content)



from typing import List, cast

import httpx

from latitude_sdk.sdk.prompts import GetAllPromptsOptions, GetPromptResult
from tests.utils import TestCase, fixtures


class TestGetAllPrompts(TestCase):
    async def test_success_global_options(self):
        endpoint = f"/projects/{self.project_id}/versions/{self.version_uuid}/documents"
        endpoint_mock = self.gateway_mock.get(endpoint).mock(
            return_value=httpx.Response(200, json=[fixtures.PROMPT_RESPONSE])
        )

        result = await self.sdk.prompts.get_all()
        request, _ = endpoint_mock.calls.last

        self.assert_requested(request, method="GET", endpoint=endpoint)
        self.assertEqual(endpoint_mock.call_count, 1)
        self.assertEqual(result, [GetPromptResult(**dict(fixtures.PROMPT))])

    async def test_success_overrides_options(self):
        options = GetAllPromptsOptions(project_id=21, version_uuid="version-uuid")
        endpoint = f"/projects/{options.project_id}/versions/{options.version_uuid}/documents"
        endpoint_mock = self.gateway_mock.get(endpoint).mock(
            return_value=httpx.Response(200, json=[fixtures.PROMPT_RESPONSE])
        )

        result = await self.sdk.prompts.get_all(options)
        request, _ = endpoint_mock.calls.last

        self.assert_requested(request, method="GET", endpoint=endpoint)
        self.assertEqual(endpoint_mock.call_count, 1)
        self.assertEqual(result, [GetPromptResult(**dict(fixtures.PROMPT))])

    async def test_success_default_version_uuid(self):
        self.sdk._options.version_uuid = None  # pyright: ignore [reportPrivateUsage]
        endpoint = f"/projects/{self.project_id}/versions/live/documents"
        endpoint_mock = self.gateway_mock.get(endpoint).mock(
            return_value=httpx.Response(200, json=[fixtures.PROMPT_RESPONSE])
        )

        result = await self.sdk.prompts.get_all()
        request, _ = endpoint_mock.calls.last

        self.assert_requested(request, method="GET", endpoint=endpoint)
        self.assertEqual(endpoint_mock.call_count, 1)
        self.assertEqual(result, [GetPromptResult(**dict(fixtures.PROMPT))])

    async def test_fails(self):
        endpoint = f"/projects/{self.project_id}/versions/{self.version_uuid}/documents"
        endpoint_mock = self.gateway_mock.get(endpoint).mock(
            return_value=httpx.Response(500, json=fixtures.ERROR_RESPONSE)
        )

        with self.assertRaisesRegex(type(fixtures.ERROR), fixtures.ERROR.message):
            await self.sdk.prompts.get_all()
        requests = cast(List[httpx.Request], [request for request, _ in endpoint_mock.calls])  # type: ignore

        [self.assert_requested(request, method="GET", endpoint=endpoint) for request in requests]
        self.assertEqual(endpoint_mock.call_count, self.internal_options.retries)
